@@ -41,6 +41,7 @@
 #include <gfarm/gfarm.h>
 #include <gfarm2fs.h>
 #include <replicate.h>
+#include <open_file.h>
 #include "gfarm2fs_msg_enums.h"
 
 /* for old interface */
@@ -207,6 +208,7 @@ static int
 gfarm2fs_getattr(const char *path, struct stat *stbuf)
 {
 	struct gfs_stat st;
+	GFS_File gf;
 	gfarm_error_t e;
 
 	e = gfs_lstat_cached(path, &st);
@@ -215,7 +217,15 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 					"gfs_lstat_cached", path, e);
 		return (-gfarm_error_to_errno(e));
 	}
-
+	if ((gf = gfarm2fs_open_file_lookup(st.st_ino)) != NULL) {
+		gfs_stat_free(&st);
+		e = gfs_pio_stat(gf, &st);
+		if (e != GFARM_ERR_NO_ERROR) {
+			gfarm2fs_check_error(GFARM_MSG_2000046, OP_GETATTR,
+				"gfs_pio_stat", path, e);
+			return (-gfarm_error_to_errno(e));
+		}
+	}
 	copy_gfs_stat(stbuf, &st);
 	gfs_stat_free(&st);
 	return (0);
@@ -280,6 +290,7 @@ gfarm2fs_readlink(const char *path, char *buf, size_t size)
 		len = size - 1;
 	memcpy(buf, src, len);
 	buf[len] = '\0';
+	free(src);
 	return (0);
 }
 
@@ -330,8 +341,7 @@ gfarm2fs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 	}
 	gfarm2fs_check_error(GFARM_MSG_2000006, OP_READDIR,
 				"gfs_readdir", path, e);
-
-	return (0);
+	return (-gfarm_error_to_errno(e));
 }
 
 static int
@@ -344,7 +354,7 @@ gfarm2fs_releasedir(const char *path, struct fuse_file_info *fi)
 	e = gfs_closedir(dp);
 	gfarm2fs_check_error(GFARM_MSG_2000007, OP_RELEASEDIR,
 				"gfs_closedir", path, e);
-	return (0);
+	return (-gfarm_error_to_errno(e));
 }
 #else /* USE_GETDIR */
 
@@ -563,7 +573,6 @@ gfarm2fs_truncate(const char *path, off_t size)
 	e2 = gfs_pio_close(gf);
 	gfarm2fs_check_error(GFARM_MSG_2000023, OP_TRUNCATE,
 				"gfs_pio_close", path, e2);
-	gfarm2fs_replicate(path, NULL);
 
 	return (-gfarm_error_to_errno(e != GFARM_ERR_NO_ERROR ? e : e2));
 }
@@ -652,6 +661,7 @@ gfarm2fs_create(const char *path, mode_t mode, struct fuse_file_info *fi)
 	}
 
 	fi->fh = (unsigned long) gf;
+	gfarm2fs_open_file_enter(gf, fi->flags|O_CREAT);
 	return (0);
 }
 
@@ -671,6 +681,7 @@ gfarm2fs_open(const char *path, struct fuse_file_info *fi)
 	}
 
 	fi->fh = (unsigned long) gf;
+	gfarm2fs_open_file_enter(gf, fi->flags);
 	return (0);
 }
 
@@ -735,19 +746,18 @@ gfarm2fs_statfs(const char *path, struct statvfs *stbuf)
 		gfarm2fs_check_error(GFARM_MSG_2000032, OP_STATFS,
 					"gfs_statfs", path, e);
 		return (-gfarm_error_to_errno(e));
-	} else {
-		stbuf->f_bsize = 1024;	/* XXX */
-		stbuf->f_frsize = 1024;	/* XXX */
-		stbuf->f_blocks = used + avail;
-		stbuf->f_bfree = avail;
-		stbuf->f_bavail = avail;
-		stbuf->f_files = files;
-		stbuf->f_ffree = -1;	/* XXX */
-		stbuf->f_favail = -1;	/* XXX */
-		stbuf->f_fsid = 298;	/* XXX */
-		stbuf->f_flag = 0;	/* XXX */
-		stbuf->f_namemax = GFS_MAXNAMLEN;
 	}
+	stbuf->f_bsize = 1024;	/* XXX */
+	stbuf->f_frsize = 1024;	/* XXX */
+	stbuf->f_blocks = used + avail;
+	stbuf->f_bfree = avail;
+	stbuf->f_bavail = avail;
+	stbuf->f_files = files;
+	stbuf->f_ffree = -1;	/* XXX */
+	stbuf->f_favail = -1;	/* XXX */
+	stbuf->f_fsid = 298;	/* XXX */
+	stbuf->f_flag = 0;	/* XXX */
+	stbuf->f_namemax = GFS_MAXNAMLEN;
 	return (0);
 }
 
@@ -757,12 +767,11 @@ gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 	gfarm_error_t e;
 
 	(void) path;
+	gfarm2fs_open_file_remove(get_filep(fi));
 	e = gfs_pio_close(get_filep(fi));
 	gfarm2fs_check_error(GFARM_MSG_2000033, OP_RELEASE,
 				"gfs_pio_close", path, e);
-	gfarm2fs_replicate(path, fi);
-
-	return (0);
+	return (-gfarm_error_to_errno(e));
 }
 
 static int
@@ -978,12 +987,20 @@ static int
 gfarm2fs_rename_cached(const char *from, const char *to)
 {
 	int rv = gfarm2fs_rename(from, to);
+	struct gfs_stat st;
 
 	if (rv == 0) {
 		gfs_stat_cache_purge(from);
 		uncache_parent(from);
 		gfs_stat_cache_purge(to);
 		uncache_parent(to);
+
+		/* try to replicate the destination file just in case */
+		if (gfs_lstat_cached(to, &st) == GFARM_ERR_NO_ERROR) {
+			if (GFARM_S_ISREG(st.st_mode))
+				gfarm2fs_replicate(to, NULL);
+			gfs_stat_free(&st);
+		}
 	}
 	return (rv);
 }
@@ -1025,6 +1042,7 @@ gfarm2fs_truncate_cached(const char *path, off_t size)
 
 	if (rv == 0)
 		gfs_stat_cache_purge(path);
+	gfarm2fs_replicate(path, NULL);
 	return (rv);
 }
 
@@ -1088,6 +1106,7 @@ gfarm2fs_release_cached(const char *path, struct fuse_file_info *fi)
 	if (rv == 0 && ((fi->flags & O_ACCMODE) == O_WRONLY ||
 			(fi->flags & O_ACCMODE) == O_RDWR))
 		gfs_stat_cache_purge(path);
+	gfarm2fs_replicate(path, fi);
 	return (rv);
 }
 
@@ -1227,9 +1246,13 @@ usage(const char *progname)
 "    -o loglevel=priority    syslog priority level (default: %s)\n"
 "    -E T                    cache timeout for gfs_stat (default: 1.0 sec.)\n"
 "    -o gfs_stat_timeout=T   same -E option\n"
-"    -o ncopy=N              number of copies (default: 0)\n"
+"    -o ncopy=N              number of copies (default: 1)\n"
 "    -o copy_limit=N         maximum number of concurrent copy creations\n"
+#ifdef HAVE_GFS_REPLICATE_FILE_TO
 "                            (default: 10)\n"
+#else /* version 2.3.X */
+"                            (default: 0)\n"
+#endif
 		"\n", progname,
 		GFARM2FS_SYSLOG_FACILITY_DEFAULT,
 		GFARM2FS_SYSLOG_PRIORITY_DEFAULT);
@@ -1299,8 +1322,12 @@ main(int argc, char *argv[])
 		.use_syslog = 1,
 		.facility = NULL,
 		.loglevel = NULL,
-		.ncopy = 0,
+		.ncopy = 1,
+#ifdef HAVE_GFS_REPLICATE_FILE_TO
 		.copy_limit = 10
+#else /* version 2.3.X */
+		.copy_limit = 0
+#endif
 	};
 
 	umask(0);
@@ -1371,6 +1398,7 @@ main(int argc, char *argv[])
 	/* end of setting params */
 
 	gfarm2fs_replicate_init(&params);
+	gfarm2fs_open_file_init();
 
 	setup_dumper();
 
